/* Copyright (C) 1992, 1993 Aladdin Enterprises.  All rights reserved.

This file is part of Ghostscript.

Ghostscript is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
to anyone for the consequences of using it or for whether it serves any
particular purpose or works at all, unless he says so in writing.  Refer
to the Ghostscript General Public License for full details.

Everyone is granted permission to copy, modify and redistribute
Ghostscript, but only under the conditions described in the Ghostscript
General Public License.  A copy of this license is supposed to have been
given to you along with Ghostscript so you can know your rights and
responsibilities.  It should be in a file named COPYING.  Among other
things, the copyright notice and this notice must be preserved on all
copies.  */

/*
 *  PIC file format devices for Ghostscript
 *  by Jonathan Hudson
 *  no rights reserved at all 
 */
#include "gdevprn.h"
#include "gsparam.h"
#include "gsstate.h"
#include "gxlum.h"

/*
 * Default X and Y resolution.
 * these are just placeholders, as we really do this when we open the
 * device
 */

#define X_DPI 60
#define Y_DPI 60

#define DEFWID 512
#define DEFHGT 660

struct gx_device_pic_s {
	gx_device_common;
	gx_prn_device_common;
	int type;
	int picmag;
};
typedef struct gx_device_pic_s gx_device_pic;

extern dev_proc_get_initial_matrix(gdev_pcl_get_initial_matrix);
private int pic_put_props (gx_device *pdev, gs_param_list *plist);
private int pic_get_props (gx_device *pdev, gs_param_list *plist);


/* The device descriptor */
private dev_proc_print_page(pic_print_page);
private dev_proc_open_device(pic_open);


#define pic_def_procs(proc_rgb, proc_color) {\
    pic_open, NULL,	NULL,\
	gdev_prn_output_page,\
	gdev_prn_close,\
	proc_rgb,\
	proc_color,\
	NULL,	/* fill_rectangle */\
	NULL,	/* tile_rectangle */\
	NULL,	/* copy_mono */\
	NULL,	/* copy_color */\
	NULL,	/* draw_line */\
	NULL,\
	pic_get_props, 	pic_put_props,\
	NULL, NULL, NULL, NULL, gx_page_device_get_page_device,\
	NULL\
	}
private dev_proc_open_device(pic_open);
private dev_proc_print_page(pic_print_page);
private dev_proc_map_rgb_color(pic4_map_rgb_color);
private dev_proc_map_color_rgb(pic4_map_color_rgb);
private dev_proc_map_rgb_color(pic8_map_rgb_color);
private dev_proc_map_color_rgb(pic8_map_color_rgb);

private gx_device_procs pic1_procs = 
   pic_def_procs(gdev_prn_map_rgb_color, gdev_prn_map_color_rgb);

private gx_device_procs pic4_procs =
  pic_def_procs(pic4_map_rgb_color, pic4_map_color_rgb);

private gx_device_procs pic8_procs =
  pic_def_procs(pic8_map_rgb_color, pic8_map_color_rgb);

gx_device_pic gs_pic8_device = 
{  
    prn_device_std_body(
	gx_device_pic,
	pic8_procs, "pic8",
	85, 110,
	X_DPI, Y_DPI,
	0,0,0,0,			/* margins */
	3, pic_print_page), 8, 504};

gx_device_pic far_data gs_pic4_device =
{
    prn_device_std_body(
	gx_device_pic,
	pic4_procs, "pic4",
	85, 110,
	X_DPI, Y_DPI,
	0,0,0,0,			/* margins */
	2, pic_print_page), 4, 504};


gx_device_pic far_data gs_pic2_device =
{
    prn_device_std_body(
	gx_device_pic,
	pic1_procs, "pic2",
	85, 110,
	X_DPI, Y_DPI,
	0,0,0,0,			/* margins */
	1, pic_print_page), 2, 504};

gx_device_pic far_data gs_pic1_device =
{
    prn_device_std_body(
	gx_device_pic,
	pic1_procs, "pic1",
	85, 110,
	X_DPI, Y_DPI,
	0,0,0,0,			/* margins */
	1, pic_print_page), 1, 504};

/* ------ Private definitions ------ */

typedef struct 
{
    short flag;
    short xpix;
    short ypix;
    short llen;
    short mode;
} PICHEAD;

#if arch_is_big_endian
# define swapword(p1) p1
#else
private ushort inline swapword (ushort val)
{
    return (ushort) (val << 8) + (val >> 8);
}
#endif


private int
pic_get_props (gx_device *pdev, gs_param_list *plist)
{
    int code;

    code = gdev_prn_get_params (pdev, plist);
    if ( code < 0) return code;
    code = param_write_int(plist, "XPix", &((gx_device_pic *)pdev)->picmag);
    return code;
}

/* Put properties. */
private int
pic_put_props (gx_device *pdev, gs_param_list *plist)
{
    int mag = ((gx_device_pic *)pdev)->picmag;
    int code;

    code = param_read_int(plist, "XPix", &mag);
    if ( code < 0 )
	return code;
    code = gdev_prn_put_params(pdev, plist);
    ((gx_device_pic *)pdev)->picmag = mag;
    if ( pdev->is_open ) gs_closedevice(pdev);
    return 0;
}

private int
pic_open(gx_device *pdev)
{
    int mag;
    short mask;
    short ar = 0;

    pdev->width = DEFWID;

    mag = ((gx_device_pic *)pdev)->picmag;

    if(mag == 0)
    {
	mag = DEFWID;
    }

    switch(((gx_device_pic *)pdev)->type)    
    {
    default:
	case 0:
      case 4:
	pdev->color_info.max_color = 3;
      case 1:
      case 2:
	mask = 7;
	break;
      case 8:
	pdev->color_info.max_color = 7;
	mask = 3;
	break;

    }
    
    if (mag != DEFWID)
    {
	if(mag < 0)
	{
	    ar = 1;
	    mag = -mag;
	}
	pdev->width = mag;

    }

    pdev->width = (pdev->width + mask) & ~(mask);

    if(pdev->width > 32000)
    {
	pdev->width = 32000;
    }

    if(pdev->width < 32)
    {
	pdev->width = 32;
    }
	
    pdev->height = (DEFHGT * pdev->width)/DEFWID;
    pdev->x_pixels_per_inch = (X_DPI * pdev->width)/DEFWID;
    pdev->y_pixels_per_inch = (Y_DPI * pdev->width)/DEFWID;
    if(ar)
    {
	pdev->height = 3* pdev->height / 4;
	pdev->y_pixels_per_inch = pdev->y_pixels_per_inch *3 /4;
    }

    return gdev_prn_open(pdev);
}

private int
pic_print_page(gx_device_printer *pdev, FILE *file)
{
    static byte mask[] = {0x80, 0x40, 0x20, 0x10, 8, 4, 2, 1};
    static byte grmask[] = {0x80, 0x20, 8, 2};
    static byte bmask[] =  {0x40, 0x10, 4, 1};

    int depth=pdev->color_info.depth;
    int orig_raster = pdev->width;
    int rsize;
    int height = pdev->height;
    byte *buf, *obuf;
    int y;
    int line_size = (orig_raster+7)*depth/8;
    short dtype = ((gx_device_pic *)pdev)->type;
    PICHEAD ph;    


    rsize = ((dtype == 2) ? 1 : 2) *((orig_raster + 7) >> 3) + 4;
    rsize = (rsize + 1) & ~1;
    
    buf = (byte *)gs_malloc(line_size, 1, "pic raster  buffer");
    obuf = (byte *)gs_malloc(rsize, 1, "pic file buffer");

    ph.flag = swapword(0x4afc);
    ph.xpix = swapword(orig_raster); 
    ph.ypix = swapword(height);
    ph.llen = swapword(rsize);

    if(dtype == 8)
    {
	ph.mode = swapword(0x0800);	
    }
    else if (dtype == 2)
    {
	ph.mode = swapword(0x0200);	
    }
    else
    {
	ph.mode = 0;	
    }

    fwrite(&ph, 1, sizeof(ph), file);

    for ( y = 0; y < height; y++ )
    {
	short i,j;
	memset(buf, 0, line_size);
	memset(obuf, 0, rsize);
	gdev_prn_copy_scan_lines(pdev, y, (byte *)buf, line_size);

	switch(dtype)
	{
	  case 1:
	    for (j = i = 0; i < line_size; i++, j += 2)
	    {
		obuf[j] = obuf[j+1] = ~buf[i];
	    }
	    break;
	  case 2:
	    for (i = 0; i < line_size; i++)
	    {
		obuf[i] = ~buf[i];
	    }
	    break;
	  case 4:
	    for (j = i = 0; i < line_size; i += 8, j += 2)
	    {
		short k;
		for (k = 0; k < 8; k++)
		{
		    obuf[j] |= (buf[i + k] & 2) ? mask[k] : 0;
		    obuf[j + 1] |= (buf[i + k] & 1) ? mask[k] : 0;
		}
	    }
	    break;
	  case 8:
	    for (j = i = 0; i < line_size; i += 8, j += 2)
	    {
		short k;
		for (k = 0; k < 4; k++)
		{
		    obuf[j]   |= (buf[i + 2*k] & 4) ? grmask[k] : 0;
		    obuf[j+1] |= (buf[i + 2*k] & 2) ? grmask[k] : 0;
		    obuf[j+1] |= (buf[i + 2*k] & 1) ? bmask[k] : 0;
		}
	    }
	    
	    break;
	}
	fwrite (obuf, 1, rsize, file);
    }
    gs_free((char *)buf, line_size, 1, "pic raster buffer");
    gs_free((char *)obuf, rsize, 1, "picx file buffer");
    return 0;
}


/* ------ Internal routines ------ */

/* Map an RGB color to a index. */
private gx_color_index
pic4_map_rgb_color(gx_device *dev, ushort r, ushort g, ushort b)
{	
	gx_color_value col = 0;
	if(r)
	    col |= 1;
	if(g)
	    col |= 2;
	return col;
}

/* Map a value back to an RGB color. */
private int
pic4_map_color_rgb(gx_device *dev, gx_color_index color, ushort prgb[3])
{
	prgb[0] = (color & 1) ? 1 : 0;
	prgb[1] = (color & 2) ? 1 : 0;
	prgb[2] = 0;
	return 0;
}

/* Map an RGB color to a index. */
private gx_color_index
pic8_map_rgb_color(gx_device *dev, ushort r, ushort g, ushort b)
{
	gx_color_value col = 0;
	if(r)
	    col |= 2;
	if(g)
	    col |= 4;
	if(b)
	    col |= 1;
	return col;
}

/* Map a value back to an RGB color. */
private int
pic8_map_color_rgb(gx_device *dev, gx_color_index color, ushort prgb[3])
{
	prgb[0] = (color & 1) ? 2 : 0;
	prgb[1] = (color & 2) ? 2 : 0;
	prgb[2] = (color & 4) ? 2 : 0;
	return 0;
}
